import random
import struct
from cPickle import loads, dumps

##############################################################################
# Network stuff

(REQUEST_INFO,
 REQUEST_PLAY,
 NOTIFY_GAMEOVER,
 ACK_GAMEOVER,
 
 RESPONSE_TAKE,
 RESPONSE_PASS
 ) = range(6)


class CommSocket(object):

    def __init__(self, socket):
        self._socket = socket

    def close(self):
        self._socket.close()

    def my_recv(self, msg_len):
        bufs = []
        total_so_far = 0
        while total_so_far < msg_len:
            chunk = self._socket.recv(msg_len - total_so_far)
            if chunk == '':
                raise RuntimeError('connection broken')
            bufs.append(chunk)
            total_so_far += len(chunk)
        return ''.join(bufs)        

    def recv_message_size(self):
        sz_str = self.my_recv(4)
        return struct.unpack('i', sz_str)[0]

    def send_message_size(self, msg):
        x = struct.pack('i', len(msg))
        self._socket.sendall(x)

    def recv_object(self):
        packet_size = 4096
        def recv_big(size):
            bufs = []
            while size > 0:
                bufsize = min(packet_size, packet_size)
                bufs.append(self._socket.recv(bufsize))
                size -= bufsize
            return ''.join(bufs)
        sz = self.recv_message_size()
        return loads(self.my_recv(sz))

    def send_object(self, obj):
        msg = dumps(obj)
        self.send_message_size(msg)
        self._socket.sendall(msg)


##############################################################################
# Game state

class GameOver(Exception):
    pass

class HiddenState(object):

    def init_deck(self):
        cards = []
        while len(cards) < 24:
            new_card = random.randint(3, 35)
            if new_card not in cards:
                cards.append(new_card)
        self.deck = cards

    def __init__(self):
        self.init_deck()


class PlayerState(object):

    def __init__(self):
        self.cards = []
        self.chips = 11
        self.name = ""
        self._score = -11
        self._sorted_cards = []
        
    def score(self):
        if self._score is not None:
            return self._score
        
        def sequences():
            x = sorted(self.cards)
            self._sorted_cards = x
            result = []
            if x == []:
                return result
            current_card = x[0]
            current_sequence = [current_card]
            for i in xrange(1,len(x)):
                new_card = x[i]
                if new_card == current_card + 1:
                    current_sequence.append(new_card)
                else:
                    result.append(current_sequence)
                    current_sequence = [new_card]
                current_card = new_card
            result.append(current_sequence)
            return result
        seqs = sequences()
        self._score = sum([s[0] for s in seqs]) - self.chips
        return self._score

    def output(self, f):
        f.write('    chips: %2d score: % -3d cards:' %
                (self.chips,
                 self.score()))
        for c in self._sorted_cards:
            f.write(' %2d' % c)
        f.write('\n')

    def add_card(self, card, chips):
        self.cards.append(card)
        self.chips += chips
        self._score = None

    def pay_chip(self):
        if self.chips == 0:
            raise RuntimeError("Invalid operation")
        self.chips -= 1
        if self._score is not None:
            self._score += 1        

class KnownState(object):

    def __init__(self, player_count):
        self.current_card = -1
        self.current_chips = 0
        self.current_player = 0
        self.players = [PlayerState() for x in xrange(player_count)]

    def take_card(self):
        self.players[self.current_player].add_card(self.current_card,
                                                   self.current_chips)
        self.current_chips = 0
        self.current_card = -1
        self.current_player = (self.current_player + 1) % len(self.players)

    def skip_card(self):
        self.players[self.current_player].pay_chip()
        self.current_chips += 1
        self.current_player = (self.current_player + 1) % len(self.players)

    def output(self, f):
        f.write('Players:\n')
        for player, i in zip(self.players, xrange(len(self.players))):
            f.write('  Player %s: %s\n' % (i, player.name))
            player.output(f)
        if self.current_card != -1:
            f.write('Current card: %2d\n' % self.current_card)
            f.write('Current chips: %2d\n' % self.current_chips)
        else:
            f.write('No card on table.\n')

class GameState(object):

    def __init__(self, player_count):
        self.hidden_state = HiddenState()
        self.known_state = KnownState(player_count)

    def take_card(self):
        self.known_state.take_card()

    def skip_card(self):
        self.known_state.skip_card()

    def flip_card(self):
        if self.hidden_state.deck == []:
            raise GameOver
        self.known_state.current_card = self.hidden_state.deck[-1]
        self.hidden_state.deck = self.hidden_state.deck[:-1]

    def check_consistency(self):
        """Checks state for consistency."""
        # Checks card count
        cards = {}
        total_cards = 0
        total_chips = 0
        for player in self.known_state.players:
            for card in player.cards:
                cards[card] = cards.get(card, 0) + 1
                assert cards[card] == 1
                total_cards += 1
            total_chips += player.chips
        assert total_chips == len(self.players)
        if self.known_state.current_card == -1:
            assert total_cards == 24
        else:
            assert total_cards == 23
